# RPC 远程调用

## 说明

此文档只适用于 jboot v3.1.0 以上，之前的版本请参考 [这里](./rpc2.x.md) 。

## 目录

- 添加依赖
- 配置
- 开始使用
- 高级功能

## 添加依赖

Jboot 支持 dubbo 和 motan，假设我们需要使用 dubbo 作为底层的 RPC 框架，需要添加如下依赖：

```
<dependency>
    <groupId>org.apache.dubbo</groupId>
    <artifactId>dubbo</artifactId>
    <version>${dubbo.version}</version>
    <exclusions>
        <exclusion>
            <groupId>org.springframework</groupId>
            <artifactId>spring-context</artifactId>
        </exclusion>
        <exclusion>
            <groupId>com.alibaba.spring</groupId>
            <artifactId>spring-context-support</artifactId>
        </exclusion>
    </exclusions>
</dependency>

<dependency>
    <groupId>org.apache.dubbo</groupId>
    <artifactId>dubbo-dependencies-zookeeper</artifactId>
    <version>${dubbo.version}</version>
    <type>pom</type>
</dependency>
```

如果使用 motan，需要添加如下依赖：

```
<dependency>
    <groupId>com.weibo</groupId>
    <artifactId>motan-core</artifactId>
    <version>${motan.version}</version>
</dependency>

<dependency>
    <groupId>com.weibo</groupId>
    <artifactId>motan-transport-netty4</artifactId>
    <version>${motan.version}</version>
</dependency>

<!-- 以下的 consul 和 zookeeper 只需要一个即可，使用哪个注册中心就导入哪个依赖  -->
<dependency>
    <groupId>com.weibo</groupId>
    <artifactId>motan-registry-consul</artifactId>
    <version>${motan.version}</version>
</dependency>

<dependency>
    <groupId>com.weibo</groupId>
    <artifactId>motan-registry-zookeeper</artifactId>
    <version>${motan.version}</version>
</dependency>
```

## 配置
在 Jboot 中，默认实现了对 Dubbo、motan 的 RPC 调用支持。在使用 RPC 远程调用之前，需要做一些基本的配置。


例如 ：
```
jboot.rpc.type = dubbo
jboot.rpc.urls = com.yourdomain.AAAService:127.0.0.1:8080,com.yourdomain.XXXService:127.0.0.1:8080
jboot.rpc.providers = com.yourdomain.AAAService:providerName,com.yourdomain.XXXService:providerName
jboot.rpc.consumers = com.yourdomain.AAAService:consumerName,com.yourdomain.XXXService:consumerName
jboot.rpc.autoExportEnable = true
```

- jboot.rpc.type ： RPC 的类型，目前只支持 dubbo 和 motan
- jboot.rpc.urls ： 一般不用配置，只有直连模式下才会去配置，此处是配置 Service接口和URL地址的映射关系。
- jboot.rpc.providers ： 配置 Service 和 Provider 的映射关系（ Motan下配置的是 Service 和 BasicService 的映射关系）。
- jboot.rpc.consumers ： 配置 Reference 和 consumer 的映射关系（ Motan下配置的是 Referer 和 BaseReferer 的映射关系）。
- jboot.rpc.autoExportEnable ： 当 Jboot 启动的时候，是否自动暴露 @RPCBean 注解的接口。


## 开始使用

一般情况下，RPC 调用需要以下几个步骤：

- 1、定义接口
- 2、编写实现类
- 3、启动 Server（Provider） 暴露服务
- 4、启动客户端、通过 RPC 调用 Server 提供的服务


**定义接口**

```java

public interface BlogService {

    public String findById();
    public List<String> findAll();
}
```

**编写实现类**

```java


@RPCBean
public class BlogServiceProvider implements BlogService {

    @Override
    public String findById() {
        return "id from provider";
    }

    @Override
    public List<String> findAll() {
        return Lists.newArrayList("item1","item2");
    }
}
```

**启动 Server 暴露服务**

```java

public class DubboServer {

    public static void main(String[] args)  {
        
        JbootApplication.setBootArg("undertow.port", "9998");

        JbootApplication.setBootArg("jboot.rpc.type", "dubbo");

        // 开启 @RPCBean 自动暴露功能，默认情况下是开启的，无需配置，
        // 但是此测试代码的 jboot.properties 文件关闭了，这里需要开启下
        JbootApplication.setBootArg("jboot.rpc.autoExportEnable", true);

        //dubbo 的通信协议配置
        JbootApplication.setBootArg("jboot.rpc.dubbo.protocol.name", "dubbo");
        JbootApplication.setBootArg("jboot.rpc.dubbo.protocol.port", "28080");


        // dubbo 注册中心的配置，
        // 当不配置注册中心的时候，默认此服务只提供了直联模式的请求
        // JbootApplication.setBootArg("jboot.rpc.dubbo.registry.protocol", "zookeeper");
        // JbootApplication.setBootArg("jboot.rpc.dubbo.registry.address", "127.0.0.1:2181");


        JbootApplication.run(args);
        System.out.println("DubboServer started...");

    }
}
```

>备注：以上的 JbootApplication.setBootArg() 里设置的内容，都可以配置到 jboot.properties 里。


**启动客户端、通过 RPC 调用 Server 提供的服务**
```java

@RequestMapping("/dubbo")
public class DubboClient extends JbootController{

    public static void main(String[] args)  {

       //Undertow端口号配置
       JbootApplication.setBootArg("undertow.port", "9999");

       //RPC配置
       JbootApplication.setBootArg("jboot.rpc.type", "dubbo");

       //设置直连模式，方便调试，默认为注册中心
       JbootApplication.setBootArg("jboot.rpc.urls", "io.jboot.test.rpc.commons.BlogService:127.0.0.1:28080");

        JbootApplication.run(args);
    }


    @RPCInject
    private  BlogService blogService;

    public void index() {
        System.out.println(blogService);
        renderText("blogId : " + blogService.findById());
    }
}
```


## 高级功能

### 更多的 dubbo 配置

目前，jboot 不支持通过 jboot.properties 直接对 service 和 reference 配置，但是可以对 provider 和 consumer 的配置，
在通过 @RPCInject 来再次复制给 service  或者 reference，结果也等同于对 service 和 reference 配置进行配置了。


#### application

http://dubbo.apache.org/zh-cn/docs/user/references/xml/dubbo-application.html

对应的 jboot 的配置前缀为： `jboot.rpc.dubbo.application`

例如：

```
jboot.rpc.dubbo.application.name = xx.xx.xx
jboot.rpc.dubbo.application.version = xx.xx.xx
```


#### registry

http://dubbo.apache.org/zh-cn/docs/user/references/xml/dubbo-registry.html


对应的 jboot 的配置前缀为： `jboot.rpc.dubbo.registry`

例如：

```
jboot.rpc.dubbo.registry.address = xx.xx.xx
```



**多 registry （多注册中心）**

默认的注册中心配置内容如下：

```
jboot.rpc.dubbo.registry.address = xx.xx.xx
jboot.rpc.dubbo.registry.port = xx.xx.xx
jboot.rpc.dubbo.registry.username = xx.xx.xx
jboot.rpc.dubbo.registry.password = xx.xx.xx
```

多注册中心可以配置如下（多 protocol 、多 consumer、多provider 都是同理）：


```
jboot.rpc.dubbo.registry.address = xx.xx.xx
jboot.rpc.dubbo.registry.port = xx.xx.xx
jboot.rpc.dubbo.registry.username = xx.xx.xx
jboot.rpc.dubbo.registry.password = xx.xx.xx


jboot.rpc.dubbo.registry.other1.address = xx.xx.xx
jboot.rpc.dubbo.registry.other1.port = xx.xx.xx
jboot.rpc.dubbo.registry.other1.username = xx.xx.xx
jboot.rpc.dubbo.registry.other1.password = xx.xx.xx


jboot.rpc.dubbo.registry.other2.address = xx.xx.xx
jboot.rpc.dubbo.registry.other2.port = xx.xx.xx
jboot.rpc.dubbo.registry.other2.username = xx.xx.xx
jboot.rpc.dubbo.registry.other2.password = xx.xx.xx
```

这样，在系统中就存在了多个注册中心，名称（name 或者 id）分别为 default（没有name的时候，默认为default）、other1、other2，这样，当一个服务（或者接口）需要在
多个注册中心暴露的时候，只需要在其 registry 配置相应的 id 即可。

例如：

```
jboot.rpc.dubbo.provider.address = default,other1
```

若当服务没有指定注册中心，那么该服务会在这三个注册中心同时暴露。


#### protocol

http://dubbo.apache.org/zh-cn/docs/user/references/xml/dubbo-protocol.html


对应的 jboot 的配置前缀为： `jboot.rpc.dubbo.protocol`

例如：

```
jboot.rpc.dubbo.protocol.host = 127.0.0.1
jboot.rpc.dubbo.protocol.port = 28080
```



#### provider

http://dubbo.apache.org/zh-cn/docs/user/references/xml/dubbo-provider.html


对应的 jboot 的配置前缀为： `jboot.rpc.dubbo.provider`

例如：

```
jboot.rpc.dubbo.provider.host = 127.0.0.1
```



#### consumer

http://dubbo.apache.org/zh-cn/docs/user/references/xml/dubbo-consumer.html


对应的 jboot 的配置前缀为： `jboot.rpc.dubbo.consumer`

例如：

```
jboot.rpc.dubbo.consumer.timeout = 127.0.0.1
```

